/*
 * A small and fast replacement for inetd
 * 2002 by Thomas Knop <tknop@maxrelax.de>
 */
#include "config.h"

#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <buffer.h>
#include <scan.h>
#include <ip4.h>
#include <fmt.h>
#include <socket.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <errno.h>
#include <string.h>
#ifdef USE_SYSLOG
#include <sys/syslog.h>
#endif

#include "sig.h"
char *prog;
int dropname = 0;
int loglevel = 1;
unsigned long maxchildren = 100;
unsigned long backlog = 20;
unsigned long uid = 0, gid = 0;

unsigned long numchildren = 0;
char localip[4];
uint16 localport;
char remoteip[4];
uint16 remoteport;

#define ENVPLUS 5
char envplus[ENVPLUS][15+IP4_FMT]; /* strlen("TCPREMOTEPORT= ") == 15 */
char *localipstr;
char *localportstr;
char *remoteipstr;
char *remoteportstr;

// ndelay_off is imlemented in libowfat 0.14, but unix.h is missing //
extern int ndelay_off(int);

#define buffer_puts2(b,s1,s2) buffer_puts3((b),(s1),(s2),0)

void buffer_puts3(buffer *b, const char *s1, const char *s2, const char *s3)
{
  if (s1) buffer_puts(b, s1);
  if (s2) buffer_puts(b, s2);
  if (s3) buffer_puts(b, s3);
}

#define log(msg,s1) log3((msg),s1,0,0)
#define log2(msg,s1,s2) log3((msg),(s1),(s2),0)

void log3(const char *msg, const char *s1, const char *s2, const char *s3)
{
#ifdef USE_SYSLOG
  syslog(errno ? LOG_ERR : LOG_INFO, 
         "%s%s%s%s", msg, s1 ? s1 : "", s2 ? s2 : "", s3 ? s3 : "");
#else
  if (!dropname) buffer_puts2(buffer_2, prog, ": ");
  buffer_puts3(buffer_2, msg, s1, s2);
  buffer_puts2(buffer_2, s3, "\n");
  buffer_flush(buffer_2);
#endif
}

void log_status(void)
{
  if (loglevel == 2) {
    char num[FMT_ULONG], max[FMT_ULONG];
    num[fmt_uint(num, numchildren)] = '\0';
    max[fmt_uint(max, maxchildren)] = '\0';
    log3("connections ", num, "/", max);
  }
}
  
#define die(err) { errno=0; die_sys(err); }

void die_sys(char *err)
{ 
#ifdef USE_STRERR
  char *s1, *s2;
  if (errno) {
    s1 = ": ";
    s2 = strerror(errno);
  } else {
    s1 = s2 = 0;
  }
  log3("fatal: ", err, s1, s2);
#else 
  char str[FMT_LONG+2];
  if (errno) {
    strncpy(str, ": ", 2);
    str[2+fmt_long(str+2, errno)] = '\0';
  } else {
    *str = '\0';
  }
  log2("fatal: ", err, str);
#endif
  exit(1);
}

int fd_copy(int to, int from)
{
  if (to == from) return 0;
  if (fcntl(from,F_GETFL,0) == -1) return -1;
  close(to);
  if (fcntl(from,F_DUPFD,to) == -1) return -1;
  return 0;
}

int fd_move(int to, int from)
{
  if (to == from) return 0;
  if (fd_copy(to,from) == -1) return -1;
  close(from);
  return 0;
}

void sigterm()
{
  _exit(0);
}

void sigchld()
{
  int pid, wstat;

  while ((pid = wait3(&wstat,WNOHANG,(struct rusage *) 0)) > 0) {
    if (loglevel == 2) {
      char pidstr[FMT_ULONG], status[FMT_LONG];
      pidstr[fmt_ulong(pidstr, pid)] = '\0';
      status[fmt_ulong(status, wstat)] = '\0';
      log3("end pid ", pidstr," status ", status);
    }
    if (numchildren) numchildren--;
    log_status();
  }
}

void exec_env(char **argv)
{
  char **env;
  register int i, j;

  for(i=0; environ[i]; i++) ;
  env = (char **)malloc((i + ENVPLUS + 1) * sizeof(char *));
  if (!env) die("no memory");
  for (i = j = 0; environ[i]; )
    env[j++] = environ[i++];
  for (i = 0; i < ENVPLUS; )
    env[j++] = envplus[i++];

  execve(*argv,argv,env);
  free(env);
}

void do_loop(int socket, char **argv)
{
  int fd;

  localportstr[fmt_ulong(localportstr, localport)] = '\0';
  for (;;) {
    while (numchildren >= maxchildren) 
      sig_pause();

    sig_unblock(SIGCHLD);
    fd = socket_accept4(socket, remoteip, &remoteport);
    sig_block(SIGCHLD);
    if (fd == -1) continue;

    numchildren++; 
    log_status();
 
    switch(fork()) {
    case 0:
      close(socket);
      if (socket_local4(fd, localip, &localport) == -1)
	die_sys("unable to get local address");
      setsockopt(socket, IPPROTO_IP, 1,(char *)0, 0); /* 1 == IP_OPTIONS */
//	/* if (!flagdelay) */ {
//	  int opt = 1; /* 1 == TCP_NODELAY */
//	  setsockopt(socket, IPPROTO_TCP, 1, &opt, sizeof opt); 
//	}
      remoteipstr[ip4_fmt(remoteipstr, remoteip)] = '\0';
      remoteportstr[fmt_ulong(remoteportstr, remoteport)] = '\0';
      localipstr[ip4_fmt(localipstr, localip)] = '\0';
      if (loglevel == 2) {
#ifdef USE_SYSLOG
        syslog(LOG_INFO, "pid %d from %s,%s to %s,%s",
               getpid(), remoteipstr, remoteportstr, localipstr, localportstr);
#else
	char pidstr[FMT_ULONG];
	pidstr[fmt_ulong(pidstr, getpid())] = '\0';
        if (!dropname) buffer_puts2(buffer_2, prog, ": ");
	buffer_puts3(buffer_2, "pid ", pidstr, " from ");
	buffer_puts3(buffer_2, remoteipstr, ",", remoteportstr);
	buffer_puts3(buffer_2, " to ", localipstr, ",");
	buffer_puts2(buffer_2, localportstr, "\n");
	buffer_flush(buffer_2);
#endif
      }
      if ((fd_move(0,fd) == -1) || (fd_copy(1,0) == -1))
	die_sys("unable to set up descriptors");
      sig_set(SIGCHLD, SIG_DFL);
      sig_unblock(SIGCHLD);
      sig_set(SIGTERM, SIG_DFL);
      sig_set(SIGPIPE, SIG_DFL);
      exec_env(argv);
      log("unable to start ", *argv);
      exit(1);
    case -1:
      log("unable to fork", (char *)0);
      --numchildren; 
      log_status();
    }
    close(fd);
  }
}

void usage(char *err)
{
  if (err) buffer_puts3(buffer_2, "error: ", err, "\n");
  buffer_puts3(buffer_2, "usage: ", prog,"\
 [-dhqQv] [-c numclients] [-b backlog] [-g gid] [-u uid] ip port programm\n");
  buffer_flush(buffer_2);
  exit(1);
}

int main(int argc, char *argv[])
{
  register char *s;
  int c, socket;

  for (s=prog=argv[0]; *s != '\0';)
    if (*s++ == '/') prog = s;

  while ((c=getopt(argc, argv, "bdhqQvc:u:g:")) != -1) {
    switch (c) {
    case 'b': scan_ulong(optarg,&backlog); break;
    case 'c': scan_ulong(optarg,&maxchildren); break;
    case 'd': dropname = 1; break;
    case 'q': loglevel = 0; break;
    case 'Q': loglevel = 1; break;
    case 'v': loglevel = 2; break;
    case 'u': scan_ulong(optarg, &uid); break;
    case 'g': scan_ulong(optarg, &gid); break;
    case 'h':
    default:  usage((char *)0);
    }
  }
  argc -= optind;
  argv += optind;

  if (!*argv) usage((char *)0); /* missing local ip */
  s = *argv++;
  if(!strcmp(s, "0")) s = "0.0.0.0";
  if (s[scan_ip4(s, localip)])
    usage("invalid local ip");

  if (!*argv) usage((char *)0); /* missong port */
  s = *argv++;
  if (s[scan_int(s, &c)] || (c < 0) || (c > 65535)) {
    usage("invalid port"); /* invalid port number */
  } else {
    localport = c;
  }

  if(!*argv) usage((char *)0); /* missing programm */

#ifdef USE_SYSLOG
  openlog(prog, 0, LOG_DAEMON);
#endif

  /* open socket */
  socket = socket_tcp();
  if (socket == -1)
    die_sys("unable to create socket");
  if (socket_bind4_reuse(socket, localip, localport) == -1)
    die_sys("unable to bind");
  if (socket_local4(socket, localip, &localport) == -1)
    die_sys("unable to get local address");
  if (socket_listen(socket, backlog) == -1)
    die_sys("unable to listen");
  ndelay_off(socket);

  /* give up rights */
  if (gid && (setgid(gid) == -1))
    die_sys("unable to set gid");
  if (uid && (setuid(uid) == -1))
    die_sys("unable to set uid");

  /* initialize plus environment */
  strcpy(envplus[0], "PROTO=TCP");
  strcpy(envplus[1], "TCPLOCALIP=");
  localipstr = envplus[1]+strlen("TCPLOCALIP=");
  strcpy(envplus[2], "TCPLOCALPORT=");
  localportstr = envplus[2]+strlen("TCPLOCALPORT=");
  strcpy(envplus[3], "TCPREMOTEIP=");
  remoteipstr = envplus[3]+strlen("TCPREMOTEIP=");
  strcpy(envplus[4], "TCPREMOTEPORT=");
  remoteportstr = envplus[4]+strlen("TCPREMOTEPORT=");

  /* fetch signals */
  sig_block(SIGCHLD);
  sig_set(SIGCHLD, sigchld);
  sig_set(SIGTERM, sigterm);
  sig_set(SIGPIPE, SIG_IGN);
  
  do_loop(socket, argv);
  return 0;
}
